/*
 * SimAnnealingBasicSolver.java
 *
 * Created on 9. nor 2006, 11:10
 *
 */

package savi.alg;

import java.lang.Math.*;

/**
 * First instance of SimAnnealingSkeleton, basic class for solving with simulated 
 * annealing algorithm. Solving cannot be stepped.
 * @author Premysl JIRIK
 */
public class SimAnnealingBasicSolver extends SimAnnealingSkeleton {
    int equilibSteps;
    
    /**
     * Standard constructor for simulated annealing algorithm with all needed 
     * parameters, creates a new instance of SimAnnealingBasicSolver.
     * @param s Initial state
     * @param temp Starting temperature value
     * @param finalTemp Final temperature value.
     * @param coolFactor Cooling factor
     * @param equilibSteps Number of steps in equilibrium.
     */
    public SimAnnealingBasicSolver(StateSkeleton s, double temp, double finalTemp, double coolFactor, int equilibSteps) {
        super(s, temp, finalTemp, coolFactor);
        this.equilibSteps = equilibSteps;
    }
    
    /**
     * This method is able to find and return starting temp.
     * @param s State
     * @param equilibSteps Number of steps in equilibrium.
     * @return Founded starting temperature
     */
    public static double findOptimalStartTemp(StateSkeleton s, int equilibSteps) {
        double tmp = 1e-5;
        StateSkeleton stat;
        int attempt, attemptAccepted;
        int measureSteps;
        
        measureSteps = 2000 / equilibSteps + 1;
        if (measureSteps < 2)
            measureSteps = 2;
        
        while (true) {
            attempt = 0;
            attemptAccepted = 0;
            
//            stat = (StateSkeleton)s.clone();
            for (int ms = 0; ms < measureSteps; ms++) {
                stat = (StateSkeleton)s.clone();
                for (int i = 0; i < equilibSteps; i++) {
                    StateSkeleton newState = (StateSkeleton)stat.clone();
                    newState.findNextState(tmp);
                    double cost = newState.getActualCost() - stat.getActualCost();
                    if (cost > 0) {
                        attempt++;
                        if (Math.random() < Math.exp(-cost / tmp)) {
                            stat = newState;
                            attemptAccepted++;
                        }
                    } else 
                        stat = newState;
                }
            }
            if (attempt > 50)
                if (2 * attemptAccepted > attempt)
                    return tmp;
            if (tmp > 1e6)          // "save break"
                return tmp;
            tmp *= 1.5;
        }
    }
    
    /**
     * Call this function lowers current temperature.
     */
    protected void cool() {
        temp *= coolFactor;
    }
    /**
     * Do iteration of equilibrium cycle with constant temperature.
     */
    protected void equilibrium() {
        for (int i = 0; i < equilibSteps; i++) {
            StateSkeleton newState = (StateSkeleton)state.clone();
            newState.findNextState(temp);
            double cost = newState.getActualCost() - state.getActualCost();
/* //debug            
            double xx = Math.random();
            double yy = Math.exp(-(cost / 10) / temp);
            boolean bbb = xx < Math.exp(-cost / temp);
            System.out.println(new Double(xx).toString() + "   " + new Double(yy).toString() + "   " + new Boolean(bbb).toString());
*/            
            if ((cost < 0) || (Math.random() < Math.exp(-cost / temp)))
                state = newState;
        }
    }
    /**
     * This function can be called to recognition end of algorithm run.
     * @return Return true if algorithm finished (current temperature is below final temperature).
     */
    protected boolean frozen() {
        return finalTemp >= temp;
    }
    /**
     * Finish and solve the problem.
     * @return Final state
     */
    public StateSkeleton solve() {
        while (!frozen()) {
           equilibrium();
           cool();
        }
        return state;
    }
    
}
