/*
 * Cities.java
 *
 * Created on 9. nor 2006, 13:36
 *
 */

package savi.alg.tsp;

import java.util.Vector;

/**
 * Class representing TSP (traveling salesman problem) instance.
 * @author Premysl JIRIK
 */
public class Cities {
    /**
     * number of cities
     */
    protected int count;
    /**
     * internal index for adding cities
     */
    protected int idx = 0;
    /**
     * list of cities
     */
    protected double cities[][];
    
    /**
     * Creates a new instance of Cities, initialize internal variables
     * @param count maximum count of cities
     */
    public Cities(int count) {
        this.count = count;
        cities = new double[count][2];
    }
    /**
     * Adds a city to the instance
     * @param x coordinate x
     * @param y coordinate y
     */
    public void addCity(int x, int y) {
        if (idx < count) {
            cities[idx][0] = x;
            cities[idx][1] = y;
            idx++;
        }
    }
    /**
     * Adds a city to the instance
     * @param x coordinate x
     * @param y coordinate y
     */
    public void addCity(double x, double y) {
        if (idx < count) {
            cities[idx][0] = x;
            cities[idx][1] = y;
            idx++;
        }
    }
    /**
     * Count distance between two cities
     * @param idx1 city A index
     * @param idx2 city B index
     * @return distance
     */
    public double distance(int idx1, int idx2) {
        idx1 = (idx1 + count) % count;
        idx2 = (idx2 + count) % count;
        double dx = cities[idx1][0] - cities[idx2][0];
        double dy = cities[idx1][1] - cities[idx2][1];
        return Math.sqrt(dx * dx + dy * dy);
    }
    /**
     * Return number of cities in the instance
     * @return number of cities in
     */
    public int getCount() {
        return count;
    }
    /**
     * Cretate <CODE>count</CODE> cities randomly
     * @param count number of cities
     */
    public void randomCities(int count) {
        idx = 0;
        this.count = count;
        cities = new double[count][2];
        for (int i = 0; i < count; i++)
           addCity(20 * Math.random(), 20 * Math.random());
    }
    /**
     * Generate grid of cities
     * @param x number of cities in direction x
     * @param y number of cities in direction y
     */
    public void gridCities(int x, int y) { 
        if ((x < 2) || (y < 2))
            return;
        
        Vector c = new Vector();
        
        for (int i = 0; i < y; i++)
            for (int j = 0; j < x; j++) {
                double [] d = new double[2];
                d[0] = (double)j / (double)(x - 1) * 20;
                d[1] = (double)i / (double)(y - 1) * 20;
                c.add(d);
            }
        citiesFromVector(c);
    }  
    
    /**
     * Load instance from vector
     * @param v list of cities 
     */
    public void citiesFromVector(Vector v) {
        this.count = v.size();
        cities = new double[count][2];        
        // mix it
        for (int i = 0; i < count; i++) {
            int idx = (int)(Math.random() * v.size());
            cities[i] = (double[])v.elementAt(idx);
            v.removeElementAt(idx);
        }

        double xmax = Double.NEGATIVE_INFINITY;
        double xmin = Double.POSITIVE_INFINITY;
        double ymax = Double.NEGATIVE_INFINITY;
        double ymin = Double.POSITIVE_INFINITY;
        for (int i = 0; i < count; i++) {
            if (cities[i][0] > xmax)
                xmax = cities[i][0];
            if (cities[i][0] < xmin)
                xmin = cities[i][0];
            if (cities[i][1] > ymax)
                ymax = cities[i][1];
            if (cities[i][1] < ymin)
                ymin = cities[i][1];
        }
        double addx = -xmin, addy = -ymin;
        double koefx = 20 / (xmax - xmin), koefy = 20 / (ymax - ymin);
        for (int i = 0; i < count; i++) {
            cities[i][0] = (cities[i][0] + addx) * koefx;
            cities[i][1] = (cities[i][1] + addy) * koefy;
        }
    }
    
    /**
     * Return coordinate x of city with given index
     * @param city city index
     * @return coordinate x
     */
    public double getCityX(int city) {
        return cities[city][0] / 20;
    }
    /**
     * Return coordinate y of city with given index
     * @param city city index
     * @return coordinate y
     */
    public double getCityY(int city) {
        return cities[city][1] / 20;
    }
}
