/*
 * StateSAT3.java
 *
 * Created on 2. erven 2006, 23:04
 *
 */

package savi.alg.weighsat3;

import java.util.ArrayList;
import savi.alg.StateSkeleton;
import savi.alg.sat3.*;

/**
 * State for MAX-WEIGHTED-3-SAT problem.
 * @author Premysl JIRIK
 */
public class StateWSAT3 implements StateSkeleton {
    /**
     * reference to solved formula
     */
    protected WeightedFormula   formula;
    /**
     * current variables assignment
     */
    protected boolean           varValues[];
    /**
     * current weight
     */
    protected static double     weightSatRatio = 0.9;

    
    /**
     * Creates a new instance of StateSAT3, intializes formula.
     * @param formula formula to solve
     */
    public StateWSAT3(WeightedFormula formula) {
        this.setFormula(formula);
    }

    /**
     * Find initial state.
     * @param temp starting temperature
     */
    public void findInitialState(double temp) {
        for (int i = 0; i < varValues.length; i++)
            varValues[i] = Math.random() < 0.5;
    }
    /**
     * Find next state.
     * @param temp current temperature
     */
    public void findNextState(double temp) {
        int idx = (int)(varValues.length * Math.random());
        varValues[idx] = !varValues[idx];
    }
    /**
     * Return current value of optimalization criterion.
     * @return opt. criterion value
     */
    public double getActualCost() {
        return (1 - formula.percentualSatisfact(varValues)) * weightSatRatio + 
               (1 - formula.percentualWeight(varValues)) * (1 - weightSatRatio);
/*
        double c = formula.percentualSatisfact(varValues) * weightSatRatio;
        if (this.isSatisfy())        
          c += formula.percentualWeight(varValues) * (1 - weightSatRatio);
        return 1 - c;
 */
    }
    /**
     * Return current fraction of satisfaction optimalization criterion.
     * @return fraction
     */
    public double getCostSatisfactPart() {
        return weightSatRatio * formula.percentualSatisfact(varValues);
    }
    /**
     *  Return current fraction of evaluation optimalization criterion.
     * @return fraction
     */
    public double getCostWeightPart() {
        return (double)(1.0 - weightSatRatio) * formula.percentualWeight(varValues);
    }
    /**
     * Return current weight
     * @return current weight
     */
    public double getWeight() {
        return (double)formula.satisfyWeight(varValues);
    }
    
    /**
     * Set ratio between satisfaction and evaluation variables.
     * @param ratio ratio between satisfaction and evaluation variables.
     */
    public void setWeightSatRatio(double ratio) {
        weightSatRatio = ratio;
    }
    /**
     * Return ratio between satisfaction and evaluation variables.
     * @return ratio
     */
    public double getWeightSatRatio() {
        return weightSatRatio;
    }
    /**
     * Return <CODE>true</CODE> if the formula is statisfacted.
     * @return <CODE>true</CODE> if the formula is statisfacted.
     */
    public boolean isSatisfy() {
        return formula.isSatisfact(varValues);
    }
    /**
     * Return current variables assignment, text stream of 0/1
     * @return current variables assignment
     */
    public String getReprezentation() {
        String s = "";
        for (int i = 0; i < varValues.length; i++)
            if (varValues[i]) s += "1"; else s += "0";
        return s;
    }
    /**
     * Return number of satisfied clauses.
     * @return number of satisfied clauses
     */
    public int countSatisfactClauses() {
        return formula.countSatisfactClauses(varValues);
    }
    /**
     * Creates and returns a copy of this object.
     * @return a copy of this object
     */
    public Object clone() {
        StateWSAT3 copy = null;
        try {
            copy = (StateWSAT3)super.clone();
            copy.varValues = this.varValues.clone();
        } 
        catch (CloneNotSupportedException ex) {
            ex.printStackTrace();
        }
        return copy;
    }
    /**
     * Return reference to formula.
     * @return reference to formula
     */
    public WeightedFormula getFormula() {
        return formula;
    }
    /**
     * Set given formula.
     * @param wf formula
     */
    public void setFormula(WeightedFormula wf) {
        formula = wf;
        varValues = new boolean[formula.varsCount()];
    }
    /**
     * Return current variables assignment.
     * @return variables assignment
     */
    public boolean[] getVariableValues() {
        return varValues;
    }

    /**
     * Return text representation of MAX-WEIGHTED-3-SAT instance.
     * @return array of lines
     */
    public String [] getSubmission() {
        ArrayList al = new ArrayList();

        al.add("c saved by SAVi (Simulated Annealing VIsualization applet)");
        al.add("c Premysl Jirik, 2006, master thesis");
        al.add("c CVUT FEL (www.feld.cvut.cz)");
        al.add("c");
        al.add("p cnf " + Integer.toString(formula.varsCount()) + " " + 
                Integer.toString(formula.clausesCount()));
        
        String line = "";
        for (int i = 0; i < formula.varWeights.length; i++) {
            line += " " + Integer.toString(formula.varWeights[i]);
        }
        al.add("w" + line);
        
        Clause [] c = formula.getClauses();
        for (int i = 0; i < c.length; i++) {
            al.add(c[i].getCnfRepresentation());
        }
        al.add("0");
        
        String [] lines = new String[al.size()];
        for (int i = 0; i < al.size(); i++) {
            lines[i] = (String)al.get(i);
        }
        return lines;
    }
}
