/*
 * Formula.java
 *
 * Created on 2. erven 2006, 23:04
 *
 */

package savi.alg.weighsat3;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.Vector;
import savi.alg.sat3.*;
import savi.alg.sat3.IntegerReader;


/**
 * Class representing boolean formula with weights.
 * @author Premysl JIRIK
 */
public class WeightedFormula extends Formula {
    /**
     * list of weights
     */
    protected int       varWeights[];
    /**
     * sum of all weights
     */
    protected int       totalWeight;
    
    /**
     * Creates a new instance of WeightedFormula, create random formula with weights
     * @param countClauses number of clauses
     * @param varsCount number of variables
     */
    public WeightedFormula(int countClauses, int varsCount) {
        super(countClauses, varsCount);
        
        varWeights = new int[varsCount];
        totalWeight = 0;
        for (int i = 0; i < varsCount; i++) {
            varWeights[i] = 1 + (int)(Math.random() * varsCount);
            totalWeight += varWeights[i];
        }
    }
    /**
     * Creates a new instance of WeightedFormula, read content from stream.
     * @param br input stream with buffered readed
     * @throws savi.alg.sat3.WrongInputFormulaData signals syntax error
     */
    public WeightedFormula(BufferedReader br) throws WrongInputFormulaData {
        super(br);
    }
    
    /**
     * Return fraction between weight of variables in state <CODE>true</CODE> and summary (total) weight
     * @param vars variables assignment
     * @return percentual weight
     */
    public double percentualWeight(boolean vars[]) {
        return satisfyWeight(vars) / (double)totalWeight;
    }
    /**
     * Return weight of variables in state <CODE>true</CODE>.
     * @param vars list of variables
     * @return weight
     */
    public int satisfyWeight(boolean vars[]) {
        int weight = 0;
        for (int i = 0; i < varsCount; i++) {
            if (vars[i])
                weight += varWeights[i];
        }
        return weight;
    }
    /**
     * Return total (summary) weight.
     * @return total weight
     */
    public int totalWeight() {
        return this.totalWeight;
    }
    /**
     * Return list of weights
     * @return list of weights
     */
    public int[] getWeights() {
        return varWeights;
    }
    
    /**
     * Read submission from stream.
     * @param br input stream with buffered readed
     * @throws savi.alg.sat3.WrongInputFormulaData signals syntax error
     */
    protected void readFromStream(BufferedReader br) throws WrongInputFormulaData {
        try {
            String s = null;
            String vals[] = null;
            IntegerReader ir = new IntegerReader(br);

            while ((s = br.readLine()) != null) {
                if (s.length() == 0)
                    continue;
                if (s.startsWith("c"))
                    continue;
                if (s.startsWith("p cnf")) {
                    vals = s.split(" +");
                    varsCount = Integer.parseInt(vals[2]);
                    clauses = new Clause[Integer.parseInt(vals[3])];
                    break;
                }
            }
            if (s == null)
                throw new WrongInputFormulaData();

            // weights
            varWeights = new int[varsCount];
            s = br.readLine();
            vals = s.split(" +");
            totalWeight = 0;
            for (int i = 0; i < varsCount; i++) {
                varWeights[i] = Integer.parseInt(vals[i + 1]);
                totalWeight += varWeights[i];
            }

            Vector dimacs = new Vector();
            for (int i = 0; i < clauses.length; i++) {
                dimacs.clear();
                int x = ir.readInt();
                while (x != 0) {
                    dimacs.add(x);
                    x = ir.readInt();
                }
                clauses[i] = new Clause(dimacs);
            }
        } catch (IOException ie) {
            throw new WrongInputFormulaData();
        } catch (IndexOutOfBoundsException ie) {
            throw new WrongInputFormulaData();
        }
    }
    
    /**
     * Create a copy of this object.
     * @return a copy of this object
     */
    public WeightedFormula clone() {
        WeightedFormula wf = new WeightedFormula(1, 1);
        wf.varsCount = this.varsCount;
        wf.totalWeight = this.totalWeight;
        wf.clauses = this.clauses.clone();
        wf.varWeights = this.varWeights.clone();
        return wf;
    }
}
