/*
 * Knapsack.java
 *
 * Created on 12. nor 2006, 20:47
 *
 */

package savi.visualization.addons;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.*;
import java.awt.*;
import java.net.*;
import java.text.DecimalFormat;
import java.io.*;
import savi.alg.StateSkeleton;

import savi.alg.knapsack.*;
import savi.dialogs.*;
import savi.visualization.support.VisualizationInfo;
import savi.visualization.support.SimPanel;

/**
 * Visualization add-on for Knapsack problem.
 * @author Premysl JIRIK
 */
public class Knapsack extends SimPanel {
    Items                   items;
    Label                   lbWeight, lbCost, lbItemsCount;
    TextField               tfCapacity;

    /** Creates a new instance of Knapsack */
    public Knapsack() {
        super(new ViewKnapsack(), 0.001, 0.00001, 0.95, 50);
    }
    
    /**
     * Initialize problem state.
     */
    protected void initState() {
        double wmax = Double.parseDouble(tfCapacity.getText());
        state = new StateKnapsack(items, wmax);
    }

    /**
     * Prepare and return help panel for this problem.
     * @return help panel
     */
    protected VisualizationInfo getInfoPanel() {
        VisualizationInfo vi = new VisualizationInfo();
        vi.addTab("Description", "/res/help/knapsack/knapsack01.gif");
        vi.addTab("Controls", "/res/help/knapsack/knapsack02.gif");
        vi.addTab("Generator", "/res/help/knapsack/knapsack03.gif");
        vi.addTab("Knapsack file", "/res/help/knapsack/knapsack04.gif");
        return vi;
    }
    
    /**
     * Return name of the problem.
     * @return name of the problem
     */
    public String getSimulationName() {
        return "Knapsack problem";
    }
    
    /**
     * Initialize components in the add-on
     */
    protected void initComponents() {
        super.initComponents();
        
        addEmptySpace();
        addInputButton("Random items").addActionListener(new BtRandomItems());
        addInputButton("Save instance").addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ((StateKnapsack)state).setCapacity(Double.parseDouble(tfCapacity.getText()));
                callSaveInstance();
            }
        });
        addInputButton("Load from URL").addActionListener(new btUrlLoad());
        addInputButton("Load from file").addActionListener(new BtFileChoose());
        
        tfCapacity = addInputField("Capacity", "255");

        lbCost = addOutputField("Current Cost", "");
        lbWeight = addOutputField("Current Weight", "");
        lbItemsCount = addOutputField("Items count", "");
        
        items = new Items(knapsackGenerator(40, 255, 255, 2, 0));
        
        settings.doLayout();
    }   
    
    /**
     * Update info panel (display) with given state.
     * @param s state
     */
    protected void updateDisplay(StateSkeleton s) {
        super.updateDisplay(s);
        
        StateKnapsack ks = (StateKnapsack)s;
        
        DecimalFormat df = new DecimalFormat("##0.0000");
        lbCost.setText(df.format(ks.getRealCost()));
        lbWeight.setText(df.format(ks.getWeight()));
        lbItemsCount.setText(Integer.toString(ks.getSelectedItemsCount()));
    }
    
    /**
     * Initialize algorithm for solving.
     * @return <CODE>true</CODE> if algorithm is properly initialized
     */
    protected boolean initAlg() {
        int cap;
        try {
            cap = (int)Double.parseDouble(tfCapacity.getText());
        } catch (NumberFormatException e) {
            new SaviErrorMessage("Maximum capatity must be a numeric type.");
            return false;
        }
        if (cap <= 0) {
            new SaviErrorMessage("Knapsack capacity must be greater than zero.");
            return false;
        }
        boolean ret = super.initAlg();
        if (ret) {
            ((StateKnapsack)state).setCapacity(cap);
        }
        
        return ret;
    }
    
    /**
     * Internal support function for generator, return <CODE>true</CODE> if item should
     * be accepted with knapsack generator with given parameters.
     * @param w weight
     * @param wmax generator maximum weight
     * @param granul generator granularity
     * @param exp generator exponent
     * @return <CODE>true</CODE> if item should be accepted with generator
     */
    protected boolean acceptableItem(int w, int wmax, int granul, double exp) {
        double thr = 0;
        switch (granul) {
            case  2: return true;                                               // does not matter
            case  0: thr = 1 / Math.pow((double)w, exp); break;                 // small
            case  1: thr = 1 / Math.pow((double)(wmax - w + 1), exp); break;    // big
        }
        if (thr >= Math.random())
            return true;
        return false;
    }
    /**
     * Generate submission (can fine tuned).
     * @param icount number of items in instance
     * @param wmax maximum weight of item
     * @param cmax maximum cost of item
     * @param granul granularity
     * @param exp exponent
     * @return instace (items list)
     */
    protected double [][] knapsackGenerator(int icount, int wmax, int cmax, int granul, double exp) {
        boolean [] issued = new boolean[wmax + 1];
        int idx = 0;
        int w;
        
        double [][] itms = new double[icount][2];
        
        while (idx < icount) {
            if (icount <= wmax) {
                w = (int)(Math.random() * wmax) + 1;
                if (issued[w])
                    continue;
            } else {
                w = (int)(Math.random() * wmax) + 1;
            }
            
            if (acceptableItem(w, wmax, granul, exp)) {
                itms[idx][0] = w;
                itms[idx][1] = (int)(Math.random() * cmax) + 1;
                issued[w] = true;
                idx++;
            }
	}
        return itms;
    }
    
    class WrongInputException extends Exception {
    }
    /**
     * Load submission (instance) from the textline.
     * @param sub text line with submission
     * @throws savi.visualization.addons.Knapsack.WrongInputException signalize syntax error
     */
    protected void loadSubmission(String sub) throws WrongInputException {
        String []vals = sub.split(" ");
        if (vals.length < 7)
            throw new WrongInputException();
        int wmax;
        double[][] itms;
        try {
            int itmCount = Integer.parseInt(vals[1]);
            wmax = Integer.parseInt(vals[2]);
            itms = new double[itmCount][2];
            for (int i = 0; i < itmCount; i++) {
                itms[i][0] = Integer.parseInt(vals[3 + i * 2]);
                itms[i][1] = Integer.parseInt(vals[4 + i * 2]);
            }
        } catch (NumberFormatException ex) {
            throw new WrongInputException();
        }
        tfCapacity.setText(Integer.toString(wmax));
        items.setItems(itms);
        ((StateKnapsack)state).setItems(items);
        initAlg();
        updateDisplay();
    }

    /**
     * Read instance from stream.
     * @param is input stream
     * @throws java.lang.Exception signalize error
     */
    protected void filePreprocess(InputStream is) throws Exception {
        String [] lines = new String[100];
        int linesCount = 0;

        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        while ((lines[linesCount] = br.readLine()) != null) {
            linesCount++;
            if (linesCount >= 100)
                break;
        }
        if (linesCount == 0)
            return;
        if (linesCount > 1) {
            SaviCommonDialog scd = new SaviCommonDialog("Choose submission");
            List lst;
            scd.content.add(lst = new List(20));
            for (int i = 0; i < linesCount; i++) {
                int idx = lines[i].indexOf(' ');
                if (idx != -1)
                    lst.add(lines[i].substring(0, idx));
            }
            scd.showForm();
            int idx = lst.getSelectedIndex();
            if (idx == -1)
                return;

            loadSubmission(lines[idx]);
        } else
            loadSubmission(lines[0]);
    }
    
    class BtFileChoose implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            running(false);
            FileDialog fd = new FileDialog(new Frame(""), "Load knapsack submission", FileDialog.LOAD);
            fd.setVisible(true);

            String curFile;
            if ((curFile = fd.getFile()) != null) {
                String filename = fd.getDirectory() + curFile;
                File f = new File(filename);
                try {
                    filePreprocess(new FileInputStream(f));
                } catch (FileNotFoundException ex) {
                    new SaviErrorMessage("File Not Found: " + filename);
                } catch (IOException ex) {
                    new SaviErrorMessage("IOException: " + filename);
                } catch (WrongInputException ex) {
                    new SaviErrorMessage("WrongInputException: submission contains syntax error.");
                } catch (Exception ex) {
                    new SaviErrorMessage(ex.toString());
                }
            }
        }
    }
    
    class btUrlLoad implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            running(false);
            SaviUrlDialog sud = new SaviUrlDialog("Knapsack - load submission from URL");
            if (sud.showForm()) {
                try {
                    URL url = new URL(sud.getURL());
                    URLConnection urlc = (URLConnection)url.openConnection();
                    //urlc.setRequestMethod("GET");
                    // Identify as Firefox 1.5 ;)
                    String userAgent = "Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.8) Gecko/20051111 Firefox/1.5";
                    urlc.setRequestProperty("User-Agent", userAgent );
                    // optionally turn off keep-alive
                    urlc.setRequestProperty( "Connection", "close" );
                    urlc.setAllowUserInteraction(false);
                    urlc.setDoInput(true);
                    urlc.setDoOutput(false);
                    urlc.setUseCaches(false);
                    urlc.connect();

                    long length = urlc.getContentLength(); // -1 if not available
                    InputStream is = urlc.getInputStream();

                    try {
                        filePreprocess(is);
                    } catch (WrongInputException ex) {
                        new SaviErrorMessage("WrongInputException: submission contains syntax error.");
                    } catch (Exception ex) {
                        new SaviErrorMessage(ex.toString());
                    }

                    is.close();
                }
                catch (Exception ex) { 
                    new SaviErrorMessage(ex.toString());
                }
            }
        }
    }
            
    class BtRandomItems implements ActionListener {
        int init_wmax = 100, init_cmax = 250, init_icount = 40, init_granul = 2;
        double init_exp = 0.6;
        
        public void actionPerformed(ActionEvent e) {
            running(false);
            SaviDialog sd = new SaviDialog("Knapgen - items generator");
            TextField fieldWmax, fieldCmax, fieldExp, fieldItemsCount;
            Choice fieldGranularity;
            sd.addInputRow("Items count", fieldItemsCount = new TextField("40", 4));
            sd.addInputRow("Weight max", fieldWmax = new TextField("100", 6));
            sd.addInputRow("Cost max", fieldCmax = new TextField("250", 6));
            sd.addInputRow("Exponent", fieldExp = new TextField("0.6", 6));
            sd.addInputRow("Granularity", fieldGranularity = new Choice());
            fieldGranularity.add("more small things");
            fieldGranularity.add("more big things");
            fieldGranularity.add("does not matter");
            
            fieldItemsCount.setText(Integer.toString(init_icount));
            fieldWmax.setText(Integer.toString(init_wmax));
            fieldCmax.setText(Integer.toString(init_cmax));
            fieldExp.setText(Double.toString(init_exp));
            fieldGranularity.select(init_granul);
            
            if (sd.showForm()) {
                int wmax, cmax, icount, granul;
                double exp;
                
                try {
                    wmax = (int)Double.parseDouble(fieldWmax.getText());
                    if ((wmax < 2) || (wmax > 1500))
                        throw new Exception();
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Weight max must be a numeric type between 2 and 1500.");
                    return;
                }
                try {
                    cmax = (int)Double.parseDouble(fieldCmax.getText());
                    if ((cmax < 2) || (cmax > 1500))
                        throw new Exception();
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Cost max must be a numeric type between 2 and 1500.");
                    return;
                }
                try {
                    icount = (int)Double.parseDouble(fieldItemsCount.getText());
                    if ((cmax < 5) || (cmax > 1500))
                        throw new Exception();
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Items count must be a numeric type between 5 and 1500.");
                    return;
                }
                try {
                    exp = Double.parseDouble(fieldExp.getText());
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Exponent must be a numeric type.");
                    return;
                }
                granul = fieldGranularity.getSelectedIndex();

                try {
                    items.setItems(knapsackGenerator(icount, wmax, cmax, granul, exp));
                    ((StateKnapsack)state).setItems(items);
                    tfCapacity.setText(Integer.toString(wmax));
                    initAlg();
                    updateDisplay();
                    screen.updateScene();
                    
                    init_wmax = wmax;
                    init_cmax = cmax;
                    init_icount = icount;
                    init_granul = granul;
                    init_exp = exp;
                } 
                catch (Exception ex) { 
                    new SaviErrorMessage(ex.toString());
                }
            }
        }
    }
    
}
    
