/*
 * TSP.java
 *
 * Created on 10. nor 2006, 14:07
 *
 */

package savi.visualization.addons;

import java.io.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.text.DecimalFormat;
import java.util.Vector;
import savi.alg.StateSkeleton;

import savi.alg.tsp.*;
import savi.dialogs.*;
import savi.visualization.support.VisualizationInfo;
import savi.visualization.support.SimPanel;

/**
 * Visualization add-on for TSP problem.
 * @author Premysl JIRIK
 */
public class TSP extends SimPanel {
    Cities      cities;
    Label       lbLength, lbCitiesCount;
            
    /** Creates a new instance of TSP */
    public TSP() {
        super(new ViewTSP(), 3, 0.12, 0.995, 50);
        
        screen.addMouseListener(new MouseListener() { 
            public void mouseClicked(MouseEvent e) { 
                if (e.getButton() == MouseEvent.BUTTON3)
                    stopEdit();
            }
            public void mouseEntered(MouseEvent e) { }
            public void mouseExited(MouseEvent e) { }
            public void mousePressed(MouseEvent e) { }
            public void mouseReleased(MouseEvent e) { }
        }
       );
    }
    
    /**
     * Initialize problem state.
     */
    protected void initState() {
        state = new StateTSP(cities);
    }
    
    /**
     * Prepare and return help panel for this problem.
     * @return help panel
     */
    protected VisualizationInfo getInfoPanel() {
        VisualizationInfo vi = new VisualizationInfo();
        vi.addTab("Description", "/res/help/tsp/tsp01.gif");
        vi.addTab("2OPT method", "/res/help/tsp/tsp02.gif");
        vi.addTab("Controls", "/res/help/tsp/tsp03.gif");
        return vi;
    }
    
    /**
     * Return name of the problem.
     * @return name of the problem
     */
    public String getSimulationName() {
        return "Travelling salesman problem (Euclidean TSP)";
    }    

    /**
     * Stop edit mode. Edit mode allows user to edit submission (can place own cities).
     */
    protected void stopEdit() {
        Vector lst = ((ViewTSP)screen).stopEditMode();
        if (lst.size() > 2) {
            cities.citiesFromVector(lst);
            ((StateTSP)state).setCities(cities);
        }
        initAlg();
        screen.repaint();
    }
    
    /**
     * Initialize components in the add-on.
     */
    protected void initComponents() {
        super.initComponents();

        addInputButton("Random").addActionListener(new BtRandomClick());
        addInputButton("Manually insert").addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                ((ViewTSP)screen).startEditMode();
                running(false);
                screen.repaint();
            }
        });
        addInputButton("Generate grid").addActionListener(new BtGridClick());
        addInputButton("Save instance").addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                callSaveInstance();
            }
        });
        addInputButton("Load from file").addActionListener(new BtLoadFromFile());
        
        lbLength = addOutputField("Current Length", "");
        lbCitiesCount = addOutputField("Cities count", "");

        settings.doLayout();

        cities = new Cities(1);
        cities.randomCities(50);
    }   
    
    /**
     * Update info panel (display) with given state.
     * @param s state
     */
    protected void updateDisplay(StateSkeleton s) {
        super.updateDisplay(s);
        
        StateTSP stsp = (StateTSP)s;
        
        DecimalFormat df = new DecimalFormat("##0.0000");
        lbLength.setText(df.format(stsp.getActualCost()));
        lbCitiesCount.setText(Integer.toString(stsp.getCities().getCount()));
    }
    
    class BtGridClick implements ActionListener {
        int init_grid_x = 15;
        int init_grid_y = 15;
        
        public void actionPerformed(ActionEvent e) {
            SaviDialog sd = new SaviDialog("Grid generator");
            TextField fieldWidth, fieldHeight;
            sd.addInputRow("Grid width:", fieldWidth = new TextField(4));
            sd.addInputRow("Grid height:", fieldHeight = new TextField(4));
            fieldWidth.setText(Integer.toString(init_grid_x));
            fieldHeight.setText(Integer.toString(init_grid_y));
           if (sd.showForm()) {
                int x, y = 0;
                try {
                    x = (int)Double.parseDouble(fieldWidth.getText());
                    y = (int)Double.parseDouble(fieldHeight.getText());

                    if ((x > 1) && (x <= 30) && (y > 1) && (y <= 30)) {
                        init_grid_x = x;
                        init_grid_y = y;
                        cities.gridCities(x, y);
                        ((StateTSP)state).setCities(cities);
                        running(false);
                        initAlg();
                        screen.repaint();
                    } else
                        throw new Exception();
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Width and height must be a numeric type between 1 and 30.");
                }
            }
        }
    }
    class BtRandomClick implements ActionListener {
        int initCitiesCount = 50;
        
        public void actionPerformed(ActionEvent e) {
            SaviDialog sd = new SaviDialog("Random generator");
            TextField fieldCount;
            sd.addInputRow("Cities count:", fieldCount = new TextField(4));
            fieldCount.setText(Integer.toString(initCitiesCount));
           if (sd.showForm()) {
                int cnt = 0;
                try {
                    cnt = Integer.parseInt(fieldCount.getText());
                    if ((cnt > 1) && (cnt <= 800)) {
                        initCitiesCount = cnt;
                        cities.randomCities(cnt);
                        ((StateTSP)state).setCities(cities);
                        running(false);
                        initAlg();
                        screen.repaint();
                    } else
                        throw new Exception();
                }
                catch (Exception ex) {
                    new SaviErrorMessage("Cities count must be a numeric type between 1 and 800.");
                }
            }
        }
    }

    
    /**
     * Read XML tag from stream (with bufferedreader)
     * @param br stream with bufferedreader
     * @throws java.lang.Exception signalize error
     * @return tag name
     */
    protected String xmlReadTag(BufferedReader br) throws Exception {
        String tag = "";
        int ch = br.read();
        while ((ch != '<') && (ch != -1))
            ch = br.read();
        if (ch == -1)
            throw new WrongTspInputException();
        ch = br.read();
        while ((ch != '>') && (ch != -1)) {
            tag += (char)ch;
            ch = br.read();
            if (tag.length() > 256)
                break;
        }
        if (ch == -1)
            throw new WrongTspInputException();
        return tag;
    }
    /**
     * Read value between XML tags from stream (with bufferedreader)
     * @param br stream with bufferedreader
     * @throws java.lang.Exception signalize error
     * @return readed value
     */
    protected String xmlReadValue(BufferedReader br) throws Exception {
        String val = "";
        br.mark(1);
        int ch = br.read();
        while ((ch != '<') && (ch != -1)) {
            val += (char)ch;
            br.mark(1);
            ch = br.read();
            if (val.length() > 256)
                break;
        }
        if (ch == -1)
            throw new WrongTspInputException();
        br.reset();
        return val.trim();
    }
    
    class WrongTspInputException extends Exception { }
    /**
     * Load instance of the problem from input stream.
     * @param is input stream
     * @throws java.lang.Exception signalize error
     */
    protected void filePreprocess(InputStream is) throws Exception {
        Vector citiesList = new Vector();
        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        br.mark(1024);
        String line = br.readLine();
        line = line.trim();
        if (!line.startsWith("<?xml "))
            br.reset();
        
        String tspTag = xmlReadTag(br);
        String tmp = tspTag.toUpperCase();
        if (!tmp.equals("TSP") && !tmp.equals("TSM"))
            throw new WrongTspInputException();
        
        String cityTag = xmlReadTag(br);
        tmp = cityTag.toUpperCase();
        
        while (tmp.equals("CITY") || tmp.equals("POINT")) {
            double [] pos = new double[2];
            pos[0] = Double.NaN; pos[1] = Double.NaN;

            String tag = xmlReadTag(br);            
            tmp = tag.toUpperCase();
            while (tmp.equals("X") || tmp.equals("Y")) {
                String val = xmlReadValue(br);
                if (tmp.equals("X"))
                    pos[0] = Double.parseDouble(val) * 20;
                else
                    pos[1] = Double.parseDouble(val) * 20;
            
                tmp = xmlReadTag(br);
                if (!tmp.equals("/" + tag))                   // is </x> | </y> ok?
                    throw new WrongTspInputException();
                tag = xmlReadTag(br);
                tmp = tag.toUpperCase();
            }
            if (!tag.equals("/" + cityTag))                   // is </city> ok?
                throw new WrongTspInputException();
            if (Double.isNaN(pos[0]) || Double.isNaN(pos[1]))
                throw new WrongTspInputException();
            citiesList.add(pos);
            
            cityTag = xmlReadTag(br);
            tmp = cityTag.toUpperCase();
        }
        if (!cityTag.equals("/" + tspTag))                    // is </tsp> tag ok?
            throw new WrongTspInputException();
        cities.citiesFromVector(citiesList);
        initAlg();
        updateView();
    }
    

    class BtLoadFromFile implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            FileDialog fd = new FileDialog(new Frame(""), "Load TSP instance", FileDialog.LOAD);
            fd.setVisible(true);

            String curFile;
            if ((curFile = fd.getFile()) != null) {
                String filename = fd.getDirectory() + curFile;
                File f = new File(filename);
                try {
                    filePreprocess(new FileInputStream(f));
                } catch (WrongTspInputException ex) {
                    new SaviErrorMessage("Input file has incorrect format or is corrupted!");
                } catch (FileNotFoundException ex) {
                    new SaviErrorMessage("File Not Found: " + filename);
                } catch (IOException ex) {
                    new SaviErrorMessage("IOException: " + filename);
                } catch (Exception ex) {
                    new SaviErrorMessage(ex.toString());
                }
            }
        }        
    }
    
}