/*
 * SimPanel.java
 *
 * Created on 8. kvten 2006, 11:02
 *
 */

package savi.visualization.support;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Locale;

import savi.alg.*;
import savi.alg.funcmin.*;
import savi.dialogs.SaviErrorMessage;
import savi.dialogs.SaviMessageBox;
import savi.visualization.addons.*;
import java.text.DecimalFormat;


/**
 * Main class providing settings control for simulation annealing, vizualization, 
 * graph and help.
 * @author Premysl JIRIK
 */
public abstract class SimPanel extends TaskPanel implements Runnable {
    TextField               tfFinalTemp, tfStartTemp, tfCoolFact, tfInnerLoop;      // alg. parameters
    TextField               tfSteps;
    Button                  btRun, btStep, btFinish, btGenerate, btFindTemp;
    Label                   lbTemp;
    Scrollbar               sbSpeed;
    Checkbox                chbCurrentSolution, chbBestSolution;
    StateButton             chbViewGraph, chbViewScreen, chbViewInfo;
            
    enum STATE { STATE_DEFINITION, STATE_RUNNING };
    STATE                   simState = STATE.STATE_DEFINITION;

    /**
     * panel with settings controls
     */
    protected Panel settings;

    /**
     * control environment panel
     */
    protected Panel ide;
    /**
     * signalizes running of algorithm
     */
    protected boolean       isRunning = false;
    /**
     * run thread reference
     */
    protected Thread        runThread;
    
    /**
     * state of problem
     */
    protected StateSkeleton           state = null;
    /**
     * palace for visualization panel
     */
    protected VisualizationPanel      screen;
    /**
     * panel for vizualization best solution
     */
    protected VisualizationPanel      bestScreen;
    /**
     * graph visualization panel
     */
    protected VisualizationGraph      graph;
    /**
     * help panel
     */
    protected VisualizationInfo       infoPanel;
    /**
     * algorithm
     */
    protected SimAnnealingStepSolver  alg;
    /**
     * best founded state
     */
    protected StateSkeleton           bestState;
    /**
     * signalizes vizualization visiblity of current or best state
     */
    protected boolean                 viewCurrent = false;
    
    double startTemp;
    double finalTemp;
    double coolingCoef;
    int innerLoops;
    
    /**
     * Creates a new instance of SimPanel, initialize SA starting values.
     * @param vp reference to visualization panel
     * @param startTemp starting temperature
     * @param finalTemp final temperature
     * @param coolingCoef cooling coeficient
     * @param innerLoops number of inner loops
     */
    public SimPanel(VisualizationPanel vp, double startTemp, double finalTemp,
            double coolingCoef, int innerLoops) {
        super();
        screen = vp;
        this.startTemp = startTemp;
        this.finalTemp = finalTemp;
        this.coolingCoef = coolingCoef;
        this.innerLoops = innerLoops;
        
        initComponents();
        initAlg();
    }
    
    /**
     * Initialize components.
     */
    protected void initComponents() {
        settings = new Panel();
        settings.setBounds(5, 5, 200, 18 * 22);
        settings.setBackground(Color.LIGHT_GRAY);
        settings.setLayout(new GridLayout(18, 2, 10, 3));
        
        settings.add(new Label("Start Temp"));
        Panel tp = new Panel();
        tp.setLayout(new GridLayout(1, 2, 0, 5));
        tp.add(tfStartTemp = new TextField(Double.toString(startTemp)));            // 3.0
        tp.add(btFindTemp = new Button("Find"));
        btFindTemp.addActionListener(new BtFindTemp());
        tp.doLayout();
        
        settings.add(tp);
        settings.add(new Label("Final Temp"));
        settings.add(tfFinalTemp = new TextField(Double.toString(finalTemp)));      // 0.12
        settings.add(new Label("Cooling koef"));
        settings.add(tfCoolFact = new TextField(Double.toString(coolingCoef)));     // 0.95
        settings.add(new Label("Inner loop"));
        settings.add(tfInnerLoop = new TextField(Integer.toString(innerLoops)));    // 10
        settings.add(btGenerate = new Button("Go to simulation"));
        btGenerate.setBackground(Color.GRAY);
        btGenerate.setForeground(Color.WHITE);
        
        // space between components
        settings.add(new Container());
        settings.add(new Container());
        
        // controls step
        Panel p = new Panel();
        p.setLayout(new GridLayout(1, 2, 0, 5));
        p.add(new Label("# steps"));
        p.add(tfSteps = new TextField("1"));
        p.doLayout();
        settings.add(p);
        settings.add(btStep = new Button("Step >"));
        // controls run
        p = new Panel();
        p.setLayout(new GridLayout(1, 2, 0, 0));
        p.add(new Label("Speed"));
        p.add(sbSpeed = new Scrollbar(Scrollbar.HORIZONTAL, 8, 1, 0, 14));
        sbSpeed.addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(AdjustmentEvent TheEvent) {
                if (isRunning)
                    if (runThread != null)
                        runThread.interrupt();
            }
        });
        p.doLayout();
        settings.add(p);
        settings.add(btRun = new Button("Run >>>"));
        // control finish
        settings.add(new Container());
        settings.add(btFinish = new Button("Finish >|"));
                
        settings.add(new Container());
        settings.add(new Container());
        settings.add(new Label("Current Temp"));
        settings.add(lbTemp = new Label(""));
        
        settings.doLayout();
        tp.doLayout();
        p.doLayout();
        
        Panel viewSettings = new Panel();
        viewSettings.setBounds(5, -2, 198 - 5, 2 * 22);
        viewSettings.setLayout(new GridLayout(2, 1));
        
        CheckboxGroup grp;
        ItemListener il;
        viewSettings.add(chbCurrentSolution = new Checkbox("view current solution        ", true, grp = new CheckboxGroup()));
        chbCurrentSolution.addItemListener(il = new ChbView());
        chbCurrentSolution.setSize(190, 22);
        viewSettings.add(chbBestSolution = new Checkbox("view best solution             ", false, grp));
        chbBestSolution.addItemListener(il);
        chbBestSolution.setSize(190, 22);
        viewSettings.setBackground(Color.LIGHT_GRAY);
        viewSettings.doLayout();

        
        Panel inner = new Panel();
        inner.setBounds(1, 1, 198, viewSettings.getHeight() - 5);
        inner.setBackground(Color.LIGHT_GRAY);
        inner.setLayout(null);
        inner.add(viewSettings);
        Panel outer = new Panel();
        outer.setBounds(5, 476 - 5 - (inner.getHeight() + 6 + 1), 200, inner.getHeight() + 2);
        outer.setLayout(null);
        outer.setBackground(Color.BLACK);
        outer.add(inner);
        
        ide = new Panel();
        ide.setLayout(null);
        ide.setBackground(Color.LIGHT_GRAY);
        ide.setBounds(0, 0, 210, 475);
        
        ide.add(settings);
        ide.add(outer);
        ide.setVisible(true);
        
        setLayout(null);
        setBackground(Color.WHITE);
        add(ide);
        screen.setBounds(210, 26, 420, 445);
        add(screen);
        
        try {
            bestScreen = (VisualizationPanel)(screen.getClass().newInstance());
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        bestScreen.setBounds(210, 26, 420, 445);
        add(bestScreen);
        bestScreen.setVisible(false);
                
        graph = new VisualizationGraph();
        graph.setBounds(210, 26, 420, 445);
        add(graph);
        graph.setVisible(false);
        
        infoPanel = getInfoPanel();
        infoPanel.setBounds(210, 26, 420, 445);
        add(infoPanel);
        infoPanel.setVisible(false);
        
        
        p = new Panel();
        MouseListener ml;
        p.setLayout(new FlowLayout(FlowLayout.LEFT, 1, 0));
        p.add(chbViewScreen = new StateButton("View solution"));
        chbViewScreen.setRaised(true);
        chbViewScreen.addMouseListener(new MouseListener() {
            public void mousePressed(MouseEvent e) { 
                if (e.getButton() == MouseEvent.BUTTON1)
                    activateView(1);
            }
            public void mouseEntered(MouseEvent e) { }
            public void mouseExited(MouseEvent e) { }
            public void mouseClicked(MouseEvent e) { }
            public void mouseReleased(MouseEvent e) { }   
        });
        p.add(chbViewGraph = new StateButton("Graph of opt. criterion"));
        chbViewGraph.addMouseListener(new MouseListener() {
            public void mousePressed(MouseEvent e) { 
                if (e.getButton() == MouseEvent.BUTTON1)
                    activateView(2);
            }
            public void mouseEntered(MouseEvent e) { }
            public void mouseExited(MouseEvent e) { }
            public void mouseClicked(MouseEvent e) { }
            public void mouseReleased(MouseEvent e) { }   
        });
        p.add(chbViewInfo = new StateButton("Help & info about problem"));
        chbViewInfo.addMouseListener(new MouseListener() {
            public void mousePressed(MouseEvent e) { 
                if (e.getButton() == MouseEvent.BUTTON1)
                    activateView(3);
            }
            public void mouseEntered(MouseEvent e) { }
            public void mouseExited(MouseEvent e) { }
            public void mouseClicked(MouseEvent e) { }
            public void mouseReleased(MouseEvent e) { }   
        });
        p.setBounds(210, 1, 420, 25);
        add(p);

        btStep.addActionListener(new BtStepAction());
        btFinish.addActionListener(new BtFinishAction());
        btRun.addActionListener(new BtRunClick());
        btGenerate.addActionListener(new BtStartClick());
        
        updateComponentsByState(simState);
    }
    
    /**
     * This method must be implemented. It should prepare and return help panel.
     * @return help panel (VisualizationInfo)
     */
    protected abstract VisualizationInfo getInfoPanel();
    
    /**
     * Adds output field (label).
     * @param title label caption
     * @param baseValue initial value
     * @return reference to output label
     */
    protected Label addOutputField(String title, String baseValue) {
        Label l;
        settings.add(new Label(title));
        settings.add(l = new Label(baseValue));
        return l;
    }
    
    /**
     * Adds input field (text field).
     * @param title caption
     * @param initialValue initial value
     * @return reference to created input text field.
     */
    protected TextField addInputField(String title, String initialValue) {
        TextField tf;
        settings.add(tf = new TextField(initialValue), 8);
        settings.add(new Label(title), 8);
        return tf;
    }
    
    /**
     *  Adds button into settings panel.
     * @param title caption
     * @return button reference
     */
    protected Button addInputButton(String title) {
        Button b;
        settings.add(b = new Button(title), 8);
        return b;
    }
    
    /**
     * Adds empty space for correct alignment.
     */
    protected void addEmptySpace() {
        settings.add(new Container(), 8);
    }
    
    /**
     * This must be implemented, initialize problem state.
     */
    protected abstract void initState();
    /**
     * Return internal state.
     * @return state
     */
    protected StateSkeleton getState() {
        return alg.getState();
    }
    
    /**
     * Update visualization panels (alert of state change).
     */
    protected void updateView() {
        if (chbCurrentSolution.getState()) {
            screen.setState(alg.getState());
            screen.updateScene();
        } else {
            bestScreen.setState(bestState);
            bestScreen.updateScene();
        }
        graph.updateScene();
        updateDisplay();
    }
    
    /**
     * Initialize algorithm for solving.
     * @return <CODE>true</CODE> if algorithm is properly initialized
     */
    protected boolean initAlg() {
        double stemp, ftemp, coolf;
        int innerl;
        try {
            stemp = Double.parseDouble(tfStartTemp.getText());
            ftemp = Double.parseDouble(tfFinalTemp.getText());
            coolf = Double.parseDouble(tfCoolFact.getText());
            innerl = Integer.parseInt(tfInnerLoop.getText());
        } catch (NumberFormatException e) {
            new SaviErrorMessage("Start/final temperature, cooling factor and inner loop must be a numeric type.");
            return false;
        }
        if (stemp <= 0) {
            new SaviErrorMessage("Start temperature must be greater than zero.");
            return false;
        }
        if (ftemp <= 0) {
            new SaviErrorMessage("Final temperature must be greater than zero.");
            return false;
        }
        if (ftemp >= stemp) {
            new SaviErrorMessage("Final temperature must be less than start temperature.");
            return false;
        }
        if ((coolf <= 0) || (coolf >= 1)) {
            new SaviErrorMessage("Cooling factor must be in range (0; 1).");
            return false;
        }
        if (innerl <= 0) {
            new SaviErrorMessage("Number of innerloop count must be greater than zero.");
            return false;
        }
        
        graph.initialize(stemp, ftemp, coolf);
        
        initState();
        alg = new SimAnnealingStepSolver(state, stemp, ftemp, coolf, innerl);
        screen.setState(alg.getState());
        bestState = (StateSkeleton)alg.getState().clone();
        bestScreen.setState(bestState);

        updateView();
        
        return true;
    }
    
    /**
     * Update info panel (display) with given state.
     * @param s state
     */
    protected void updateDisplay(StateSkeleton s) {
        DecimalFormat df = new DecimalFormat("##0.000000");
        lbTemp.setText(df.format(alg.getTemp()));
    }
    /**
     * Update info panel (display) in dependence of best/current state.
     */
    protected void updateDisplay() {
        if (viewCurrent)
            updateDisplay(alg.getState());
        else
            updateDisplay(bestState);
    }
    
    /**
     * Makes one step of algorithm.
     */
    protected void oneStep() {
        alg.makeStep();
        alg.makeStep();
        double curCost = alg.getState().getActualCost();
        graph.addValue(curCost);
        if (bestState.getActualCost() > curCost) {
            bestState = (StateSkeleton)alg.getState().clone();
            bestScreen.setState(bestState);
        }
    }
    /**
     * Makes one step of algorithm and updates all information components.
     */
    protected void oneStepAndUpdateView() {
        oneStep();
        updateView();
    }
    
    /**
     * Set components according to running state.
     * @param r running state
     */
    protected void running(boolean r) {
        if (r) {
            isRunning = true;
            btRun.setLabel("Stop");
        } else {
            isRunning = false;
            btRun.setLabel("Run >>>");
        }
    }
    
    /**
     * Thread is constructed using a separate Runnable <CODE>run</CODE> object. The thread
     * is used for animation.
     */
    public void run() {
        while (isRunning) {
            oneStepAndUpdateView();
            if (alg.isEnd()) {
                running(false);
                simState = STATE.STATE_DEFINITION;
                updateComponentsByState(simState);
                viewCurrent = false;
                activateScreenView();
            }
            try {
                int ms = (int)(Math.exp(Math.log(10) * sbSpeed.getValue() / 4));
                Thread.sleep(ms);
            } catch (InterruptedException e) { }
        }
    }
    
    /**
     * Starts visualization run.
     */
    public void startAnimation() {
        runThread = new Thread(this);
        runThread.start();
    }
    
    /**
     * Traverse container and set enabled to each component. It is recursive method.
     * @param c container to traverse
     * @param e enable value
     */
    protected void traverseComponentsSetEnabled(Component c, boolean e)
    {
        if (c instanceof Container) {
            Container c1 = (Container)c;
            for (int i = 0; i < c1.getComponentCount(); i++)
                traverseComponentsSetEnabled(c1.getComponent(i), e);
        }
        c.setEnabled(e);
    }
    /**
     * Updates components by current simulation state (definition problem / running simulation).
     * @param s simulation state
     */
    protected void updateComponentsByState(STATE s) {
        boolean enb = s == STATE.STATE_DEFINITION;
        int n = 0;
        while (n < settings.getComponentCount()) {
            Component c = settings.getComponent(n);
            if (c == btGenerate) {
                enb = !enb;
                n++;
                continue;
            }
            traverseComponentsSetEnabled(c, enb);
            if (c == btFinish)
                break;
            n++;
        }
        if (s == STATE.STATE_DEFINITION)
            btGenerate.setLabel("Go to simulation");
        else
            btGenerate.setLabel("Break simulation");
    }
    
    /**
     * Enables or disables the IDE (all components for input).
     * @param e enables the IDE if <CODE>true</CODE> else disable it
     */
    public void setEnableIDE(boolean e) {
        ide.setEnabled(e);
        chbViewScreen.setEnabled(e);
        chbViewGraph.setEnabled(e);
        chbViewInfo.setEnabled(e);
    }
    
    /**
     * This method switch between best and current solution. Set 
     * <CODE>viewCurrent</CODE> variable and call <CODE>activateScreenView</CODE>.
     * @param vc activates current solution view if <CODE>true</CODE> else best solution is 
     * activated
     */
    public void setViewCurrent(boolean vc) {
        viewCurrent = vc;
        activateScreenView();
    }
    
    /**
     * Activate given view.
     * 1 ... view current / best solution
     * 2 ... graph of optimalization criterion
     * 3 ... help
     * @param v view index
     */
    public void activateView(int v) {
        screen.setVisible(v == 1 && viewCurrent);
        bestScreen.setVisible(v == 1 && !viewCurrent);
        chbViewScreen.setRaised(v == 1);
        
        graph.setVisible(v == 2);
        chbViewGraph.setRaised(v == 2);
        
        infoPanel.setVisible(v == 3);
        chbViewInfo.setRaised(v == 3);
    }
    
    /**
     * This method switch between best and current solution according to 
     * <CODE>viewCurrent</CODE> variable.
     */
    public void activateScreenView() {
       screen.setVisible(viewCurrent);
       bestScreen.setVisible(!viewCurrent);
       activateView(1);
       updateView();
       chbBestSolution.setState(!viewCurrent);
       chbCurrentSolution.setState(viewCurrent);
    }
    
    /**
     * This method opens File save dialog and save current problem instance.
     */
    public void callSaveInstance() {
        FileDialog fd = new FileDialog(new Frame(""), "Save instance", FileDialog.SAVE);
        fd.setVisible(true);

        String curFile;
        if ((curFile = fd.getFile()) != null) {
            String filename = fd.getDirectory() + curFile;
            try {
                FileWriter fw = new FileWriter(new File(filename), false);
                String [] submiss = state.getSubmission();
                for (int i = 0; i < submiss.length; i++) {
                    fw.write(submiss[i]);
                    fw.write(10);
                }
                fw.close();
            } catch (FileNotFoundException ex) {
                new SaviErrorMessage("File " + filename + "cannot be created.");
            } catch (IOException ex) {
                new SaviErrorMessage("IOException: " + filename);
            } catch (Exception ex) {
                new SaviErrorMessage(ex.toString());
            }
        }
    }
    
    
    class BtStartClick implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            switch (simState) {
                case STATE_DEFINITION:
                    if (!initAlg())
                        return;
                    simState = STATE.STATE_RUNNING;
                    break;
                    
                case STATE_RUNNING:
                    running(false);
                    simState = STATE.STATE_DEFINITION;
                    break;
            }
            updateComponentsByState(simState);
        }
    }
    
    class BtFindTemp implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            int innerl;
            try {
                innerl = Integer.parseInt(tfInnerLoop.getText());
            } catch (NumberFormatException ex) {
                new SaviErrorMessage("Inner loop value must be a numeric type.");
                return;
            }
            double t = SimAnnealingBasicSolver.findOptimalStartTemp(state, innerl);
//            tfStartTemp.setText(Double.toString(t));
            tfStartTemp.setText(String.format(Locale.ENGLISH, "%6.4f", t));
        }
    }
            
    class BtRunClick implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            if (isRunning)
                running(false);
            else {
                running(true);
                startAnimation();
            }
        }
    }
    class BtStepAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            try {
                int steps = Integer.parseInt(tfSteps.getText());
                for (int i = 0; i < steps; i++)
                    oneStepAndUpdateView();
            } catch (NumberFormatException ex) {
                return;
            }
        }
    }
    class BtFinishAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            while (true) {
                oneStep();
                if (alg.isEnd()) {
                    running(false);
                    simState = STATE.STATE_DEFINITION;
                    updateComponentsByState(simState);
                    setViewCurrent(false);
                    break;
                }
            }            
            updateView();
        }
    }
    
    class ChbView implements ItemListener {
       public void itemStateChanged(ItemEvent e) {
           setViewCurrent(((e.getItem() == chbCurrentSolution.getLabel()) && (e.getStateChange() == e.SELECTED)));
       }    
    }
    
}
